"""
This is a tool to detect legal citations in text, and
generate relevant hyperlinks. For more information, see
https://raindrum.github.io/citeurl.
"""

# NOTE: the commented-out imports are imported conditionally later on

# python standard imports
from sys import stdin
from argparse import ArgumentParser, SUPPRESS
from pathlib import Path
# import webbrowser
# from tempfile import NamedTemporaryFile
# from time import sleep

# internal imports
from .citator import Citator, insert_links, _get_default_citator
from .authority import list_authorities
# from .web.server import serve, App
# from .web.makejs import makejs

def main():
    parser = ArgumentParser(
        description=__doc__,
        epilog=(
            "You must specify one of the four commands, e.g. 'citeurl "
            + "lookup'. You can find more info for each command by running "
            + "'citeurl <COMMAND> -h'."
        )
    )
    
    # common options
    shared_args = ArgumentParser(add_help=False)
    shared_args_grp = shared_args.add_argument_group(title='citator options')
    shared_args_grp.add_argument(
        '-t', '--template-file',
        metavar='FILE',
        action='append',
        default=SUPPRESS,
        help=(
            'path to a YAML file with extra citation templates for CiteURL '
            + 'to use. See https://raindrum.github.io/citeurl/template-yamls '
            + 'for help writing templates. The -t option can be used '
            + 'multiple times to load multiple files.'
        )
    )
    shared_args_grp.add_argument(
        '-n', '--no-default-templates',
        action='store_true',
        help="prevent loading CiteURL's default templates",
    )
    
    subparsers = parser.add_subparsers(
        dest='command',
        title='commands',
        required=False
    )
    
    ####################################################################
    # TEXT PROCESSOR OPTIONS
    ####################################################################
    
    process_parser = subparsers.add_parser(
        'process',
        aliases=['p'],
        parents=[shared_args],
        help='detect all citations in a block of text',
    )
    process_parser.add_argument(
        '-i',
        '--input',
        metavar='FILE',
        help='path to a file from which to read input text',
    )
    process_parser.add_argument(
        'text',
        help=(
            'a string containing legal citations. Can be omitted if -i is '
            + 'specified, or if text is piped from stdin'
        ),
        nargs='*',
        default=(None if stdin.isatty() else [''.join(stdin.read()[0:-1])])
    )
    process_parser.add_argument(
        '-r', '--no-redundant-links',
        action='store_true',
        help=(
            "Don't add a hyperlink if it would point to the same URL as the "
            "last hyperlink."
        )
    )
    process_parser.add_argument(
        '-c', '--css-class',
        metavar='CLASS',
        default='citation',
        help=(
            'the class each inserted <a> element will have. '
            + 'Defaults to "citation"'
        )
    )
    output = process_parser.add_mutually_exclusive_group(required=False)
    
    output.add_argument(
        '-o', '--output',
        metavar='FILE',
        help='write hyperlinked text to a file instead of stdout',
    )
    output.add_argument(
        '-b', '--browse',
        action='store_true',
        help=(
            'make a temporary HTML file with the hyperlinked text, '
            + 'and open it in a browser.'
        )
    )
    output.add_argument(
        '-a', '--authorities',
        metavar='NUMBER',
        action='store',
        nargs='?',
        const=-1,
        type=int,
        help=(
            'return list of all the authorities cited in the text, '
            + 'with information about each one. If a number is given, '
            + 'return only the X authorities with the most citations.'
        )
    )
    
    ####################################################################
    # LOOKUP COMMAND OPTIONS
    ####################################################################
    
    lookup_parser = subparsers.add_parser(
        'lookup',
        aliases=['l'],
        parents=[shared_args],
        help='look up the URL or other information for a single citation',
    )
    lookup_parser.add_argument(
        'text',
        help=(
            'a string containing legal citations. Can be omitted if text is '
            + ' piped from stdin'
        ),
        nargs='*',
        default=(None if stdin.isatty() else [''.join(stdin.read()[0:-1])])
    )
    lookup_parser.add_argument(
        '-b', '--browse',
        action='store_true',
        help='open the resulting URL, if any, in a browser'
    )
    lookup_parser.add_argument(
        '-s', '--strict',
        action='store_true',
        help=(
            'match templates the same way the "process" command does: '
            + "use case-sensitive regex, and ignore templates' broadRegexes"
        )
    )
    
    ####################################################################
    # HOST COMMAND OPTIONS
    ####################################################################
    
    server_parser = subparsers.add_parser(
        'host',
        aliases=['h'],
        parents=[shared_args],
        help='host a citation lookup server via HTTP',
    )
    server_parser.add_argument(
        '-s', '--serve',
        action='store_true',
        help="make the server available to your local network"
    )
    server_parser.add_argument(
        '-p', '--port',
        action='store',
        default=53037,
        type=int,
        help=('the port to use for hosting. Defaults to 53037')
    )
    
    ####################################################################
    # MAKEJS COMMAND OPTIONS
    ####################################################################
    
    makejs_parser = subparsers.add_parser(
        'makejs',
        aliases=['m'],
        parents=[shared_args],
        help='export an instance of CiteURL as JavaScript',
    )
    
    makejs_parser.add_argument(
        '-e', '--entire-page',
        action='store_true',
        help='generate a full HTML page instead of just JavaScript',
    )
    makejs_parser.add_argument(
        '-s', '--sources-table',
        action='store_true',
        help=(
            'include a table listing all the supported sources of law. '
            + 'Implies --entire-page.'
        )
    )
    makejs_parser.add_argument(
        '-o', '--output',
        metavar='FILE',
        help='write output to a file instead of stdout',
    )
    
    ####################################################################
    # SETUP
    ####################################################################
    
    args = parser.parse_args()
    if not args.command:
        parser.print_help()
        return
    
    # un-abbreviate command vars
    if args.command == 'l':
        args.command = 'lookup'
    elif args.command == 'p':
        args.command = 'process'
    elif args.command == 'h':
        args.command = 'host'
    elif args.command == 'm':
        args.command = 'makejs'
    
    # these commands share functionality for parsing input,
    # as well as the 'browse' option
    if args.command in ['process', 'lookup']:
        if args.browse:
            import webbrowser
        
        if 'input' in args and args.input:
            with open(args.input, 'r') as f:
                text = f.read()[0:-1]
        elif args.text:
            text = ' '.join(args.text)
        else:
            if args.command == 'process':
                process_parser.print_help()
            else:
                lookup_parser.print_help()
            return
        
    # create citator
    if args.no_default_templates:
        citator = Citator(defaults=None)
    else:
        citator = _get_default_citator()
    if 'template_file' in args:
        for path in args.template_file:
            citator.load_yaml(Path(path).read_text())
    if not citator.templates:
        raise SystemExit("Can't use '-n' without specifying a template file.")
    
    ####################################################################
    # TEXT PROCESSOR COMMAND
    ####################################################################
    
    if args.command == 'process':
                  
        if args.authorities:
            authorities = list_authorities(citator.list_cites(text))
            if args.authorities != -1:
                authorities = authorities[:args.authorities]
            outputs = []
            for authority in authorities:
                output = f"Authority:  {authority}"
                output += f"\nSource:     {authority.template}"
                if authority.URL:
                    output += "\nURL:        " + authority.URL
                output += f"\nReferences: {len(authority.citations)}"
                outputs.append(output)
            print('\n\n'.join(outputs))
        
        else:
            out_text = insert_links(
                text = text,
                citator = citator,
                attrs = {'class': args.css_class},
                redundant_links = not args.no_redundant_links,
            )
        
            if args.output: #write processed text to file
                with open(args.output, 'w') as f:
                    f.write(out_text)
            
            elif args.browse: # write processed text to tempfile; open it
                from tempfile import NamedTemporaryFile
                from time import sleep
                with NamedTemporaryFile('r+', suffix='.html') as f:
                    f.write(out_text)
                    webbrowser.open('file://' + f.name)
                    f.seek(0)
                    sleep(2)

            else: # just print to stdout
                print(out_text)

    ####################################################################
    # LOOKUP COMMAND
    ####################################################################
    
    elif args.command == 'lookup':
        citation = citator.cite(text, broad=False if args.strict else True)
        
        # cancel if no citation found
        if not citation:
            print("Couldn't recognize citation.")
            return
        
        # pretty-print citation information by indenting each bit of info
        # based on the length of the longest token name
        key_lengths = [
            len(k) for k in citation.tokens.keys()
            if citation.tokens[k]
        ]
        tab_width = max(key_lengths) + 2
        print('Source: '.ljust(tab_width) + str(citation.template))
        for key, value in citation.tokens.items():
            if not value:
                continue
            print(
                f"{key.replace('_', ' ').title()}: ".ljust(tab_width)
                + value
            )
        print('URL: '.ljust(tab_width) + (citation.URL or 'Unavailable'))
        
        # open the URL in a browser if applicable
        if args.browse and citation.URL:
            webbrowser.open(citation.URL)
    
    ####################################################################
    # HOST COMMAND
    ####################################################################        

    elif args.command == 'host': # host a CiteURL server
        try:
            from .web.server import serve, App
            serve(
                App(citator),
                localhost=not args.serve,
                port=args.port,
            )
        except FileNotFoundError: # ModuleNotFoundError:
            print(
                'Server features depend on flask and gevent. You can '
                + 'probably install them with the following command:'
            )
            print('python3 -m pip install flask gevent')
    
    ####################################################################
    # MAKEJS COMMAND
    #################################################################### 
    
    elif args.command == 'makejs':
        from .web.makejs import makejs
        output = makejs(
            citator,
            entire_page=args.entire_page or args.sources_table,
            include_sources_table=args.sources_table,
        )

        # save or print output
        if args.output:
            with open(args.output, 'w') as f:
                f.write(output)
        else:
            print(output)



# python standard imports
from typing import Iterable
import re

def process_pattern(
    pattern: str,
    replacements: dict[str, str],
    token_prefix: str = None,
    add_word_breaks: bool = False,
):
    """
    For a given regex pattern, find all the places that a key in the
    replacements dict appears, enclosed in curly braces. Replace each
    one with the corresponding value, enclosed in parentheses.
    
    If token_prefix is provided, it will only replace placeholders that
    start with that prefix, e.g. the 'same' in "{same volume}" or
    "{same reporter}".
    
    If add_word_breaks is True, a mandatory word break will be added at
    the beginning and end of the pattern. 
    """
    for key, value in replacements.items():
        if not value:
            continue
        if token_prefix:
            marker = '{%s %s}' % (token_prefix, key)
        else:
            marker = '{%s}' % key
        if not (value.startswith('(') and value.endswith(')')):
            value = f'({value})'
        value = fr'{value}(?=\W|$)'
        pattern = pattern.replace(marker, value)
    if add_word_breaks:
        pattern = rf'(?<!\w){pattern}(?!=\w)'
    return pattern


def match_regexes(text: str, regexes: list, span: tuple=(0,)) -> Iterable:
    """
    For a given text and set of regex Pattern objects, generate each
    non-overlapping match found for any regex. Regexes earlier in
    the list take priority over later ones, such that a span of text
    that matches the first regex cannot also match the second.
    """
    start = span[0]
    if len(span) > 1:
        end = span[1]
    else:
        end = None
    
    keep_trying = True
    while keep_trying:
        span = (start, end) if end else (start,)
        matches = []
        for regex in regexes:
            match = regex.search(text, *span)
            if match:
                matches.append(match)
        if matches:
            matches.sort(key=lambda x: (x.span()[0], -len(x.group())))
            start = matches[0].span()[1]
            yield matches[0]
        else:
            keep_trying = False



"strings and functions that are used by both server.py and makejs.py"

# python standard imports
from urllib.parse import quote_plus, urlsplit
from importlib.metadata import version
from re import sub
from pathlib import Path

########################################################################
# External Files
########################################################################

_dir = Path(__file__).parent.absolute()
CSS_PATH = _dir / 'style.css'
LOGO_PATH = _dir / 'logo.svg'

########################################################################
# HTML Templates
########################################################################

PAGE_TEMPLATE = """
<head>
  <meta content="width=device-width, initial-scale=1" name="viewport"/>
  {head}
</head>
<body>
<div class="content">
  {content}
</div>
<footer>{relation}
<a href="https://raindrum.github.io/citeurl">CiteURL</a>
{version} and subject to absolutely no warranty.</footer>
</body>"""

SOURCES_INTRO = """
<p>Below, you'll find the sources of law that this instance of CiteURL
supports. If you want to know what citation formats each one will
recognize, you can click "view regex" to see a diagram.</p>"""

SOURCES_TABLE = """
<div class="table-wrapper">
  <table>
    <thead>
      <th>Source of Law</th>
      <th>Website</th>
      <th>Citation Format</th>
    </thead>
    <tbody>
      {rows}
    </tbody>
  </table>
</div>"""

SOURCES_TABLE_ROW = """
      <tr>
        <td>{name}</td>
        <td><a href="{domain_URL}">{domain_name}</a></td>
        <td><a href="https://regexper.com#{escaped_regex}">view regex</a></td>
      </tr>"""

VERSION = f"v{version('citeurl')}"

########################################################################
# Functions
########################################################################

def format_page(
    text: str,
    js: str='',
    inline_css: bool=False,
    inline_logo: bool=False,
    relation_to_citeurl: str='Powered by',
    **kwargs
):
    """
    Returns PAGE_TEMPLATE with the given text inserted into the body.
    All placeholders (in curly braces) in the given text must be filled
    via keyword arguments. In addition to those mandatory keywords, you
    may provide any of the following:
    
    Arguments:
        js: JavaScript to be inserted directly into the page head
        inline_css: whether the CSS styling should be embedded in the
            instead of an outside link. Default: False
        inline_logo: whether the logo SVG should be embedded in the page
            instead of an outside link. Default: False
        relation_to_citeurl: a string that will be inserted into the
            page footer, just before the link to CiteURL and the
            subsequent disclaimer. Default: 'Powered by'.
        title: the HTML page title
    """
    for k, v in kwargs.items():
        text = text.replace('{' + k + '}', v)
    #text = text.format(**kwargs)
    if inline_logo:
        logo = LOGO_PATH.read_text()
    else:
        logo = '<img src="logo.svg" alt="CiteURL logo">'
    if inline_css:
        css_section = f'<style>{CSS_PATH.read_text()}</style>'
    else:
        css_section = '<link rel="stylesheet" href="style.css">'
    if not inline_logo: # add favicon to header
        favicon = '<link rel="icon" href="logo.svg">\n'
    else:
        favicon = ''
    js_section = f"<script>{js}</script>" if js else ''
    return PAGE_TEMPLATE.format(
        head=favicon + css_section + '\n' + js_section,
        content=logo + '\n' + text,
        version=VERSION,
        relation=kwargs.get('relation') or 'Powered by',
    )

def sources_table(citator):
    """
    Return the content for an HTML table listing every template that the 
    citator can link to.
    """ 
    rows = []
    for template in citator.templates.values():
        # skip templates that can't make URLs
        if not template.__dict__.get('URL_builder'):
            continue
        
        URL = urlsplit(''.join(template.URL_builder.parts))
        domain_URL = f'{URL.scheme}://{URL.netloc}'
        domain_name = URL.hostname
        regex = unify_regex(template, simplify_for_regexper=True)

        rows.append(SOURCES_TABLE_ROW.format(
            name=template.name,
            domain_URL=domain_URL,
            domain_name=domain_name,
            escaped_regex=quote_plus(regex).replace('+', '%20')
        ))
    
    return SOURCES_TABLE.format(rows=''.join(rows))

def unify_regex(template, simplify_for_regexper: bool=False):
    """
    Combine the given template's regexes (if there are multiple) into
    one long regex that matches any one of them. Wherever possible,
    insert any mandatory lookups from the template's operations, so that
    the resulting regex will be restricted to each option from the
    lookups.
    
    Arguments:
        template: the template object to get a regex for
        simplify_for_regexper: whether to strip lookbehinds, lookaheads,
            and the names of capture groups so that the resulting URL is
            compatible with regexper.com. Note that if this is false,
            duplicated capture groups will be renamed to avoid conflict.
    """
    regexes = [p.pattern for p in template.regexes.copy()]
    # append numbers to repeated capture group names to prevent conflict
    for i, regex in enumerate(regexes[1:]):
        regexes[i+1] = sub(r'\?P<(.+?)>', '?P<\1' + f'{i}>', regex)
    regex = '|'.join(regexes)
    
    # whenever a template operation uses a token for a mandatory lookup,
    # and that token hasn't been modified by another operation yet,
    # insert the lookup's regexes where the token goes, to make the
    # final regex more specific.
    for token_name, token in template.tokens.items():
        for i, edit in enumerate(token.edits):
            # only lookup operations are used here
            if edit.action is not 'lookup':
                continue
            
            # don't bother if the lookup's input token is modified before
            # the lookup
            already_modified = False
            for prior_op in token.edits[:i-1]:
                if prior_op != edit:
                    already_modified = True
                    break
            if already_modified:
                continue
        
            # modify the regex
            pattern = '\(\?P<' + token_name + '\d*>.+?(?<!\\\)\)'
            repl = '(' + '|'.join(edit.data.keys()) + ')'
            regex = sub(pattern, 'PlAcEhOlDeR122360', regex)
            regex = regex.replace('PlAcEhOlDeR122360', repl)
        
    if simplify_for_regexper:
        # remove lookaheads and lookbehinds
        regex = sub(r'\(\?(<!|<=|!|=).+?\)', '', regex)
        
        # escape slashes
        regex = regex.replace('/', r'\/')

        # remove capture group names
        regex = sub(r'\?P<.+?>', '', regex)
    
    return regex



# python standard imports
import socket
from urllib.parse import unquote, urlsplit
from re import sub
from html import escape

# internal imports
from .resources import format_page, sources_table, SOURCES_INTRO
from .. import Citator, insert_links

# third-party imports
from flask import Flask, redirect, make_response, send_file, request
# from gevent import monkey (imported in serve())
# from gevent.pywsgi import WSGIServer (imported in serve())

_APP = Flask(__name__, static_url_path='')

########################################################################
# Messages
########################################################################

INDEX_PAGE = """
<h1>{name}</h1><title>{name}</title>
<div class="narrow">
<p>Type a legal citation into the box below, and I'll try to send you
directly to the case or law that it references:</p>
<form class="searchbar" method="get">
  <input type="search" required name="s" placeholder="Enter citation..."
  maxlength=400 label="citation search bar"><button type="submit">Go</button>
</form>
<p>This web app supports citations to <a href="sources">various sources</a>
of law. You can also use it to <a href="parser">detect citations</a> in a
longer text, like a court opinion.</p>
</div>
"""

SOURCES_PAGE = """
<h1>Sources of Law</h1><title>Sources of Law</title>
{intro}
{table}
<p><a href="/"><button>Go Back</button></a></p>
"""

PARSER_PAGE = """
<h1>Text Parser</h1><title>Text Parser</title>
<p>Paste some text into the box below and click "Parse" to process
the text and find every <a href="sources">supported citation</a>
it contains.</p> 
<form action="parser#output" method="post">
<textarea required cols=80 rows=16 maxlength=400000 name="text"
placeholder="Paste text here.." label="Input text for
parser">{given_text}</textarea>
<p><a href="/"><button type="button">Go Back</button></a>
<button type="submit">Parse</button></p>
</form>
{output}
"""

# Errors

ERROR_400 = """
<h1>Unknown Citation</h1><title>Unknown Citation</title>
<p>Sorry, "{query}" isn't a citation I recognize.</p>
<p><a href="/"><button>Go Back</button></a></p>
"""

ERROR_501 = """
<h1>Missing URL</h1><title>Missing URL</title>
<p>Sorry, I can tell that's a {template} citation but I don't have a
link for it.</p>
<p><a href="/"><button>Go Back</button></a></p>
"""

ERROR_413 = """
<h1>Query Too Long</h1><title>Query Too Long</title>
<p>Sorry, that's too many characters for the server to process.</p>
<a href="/"><button>Go Back</button></a>
"""

########################################################################
# Routes
########################################################################

@_APP.route('/<query>')
def _handle_query(query: str):
    """
    Try to interpret query as a citation and return a 301 redirect if
    it returns a URL. Otherwise return 501 error if it matches a
    citation without a URL, or 400 if no match.
    """
    if len(query) > _APP.max_chars:
        return format_page(ERROR_501)
    
    query = escape(unquote(query))
    cite = _APP.citator.cite(query)
    if cite:
        if cite.URL:
            return redirect(cite.URL, code=301)
        else:
            return make_response(
                format_page(ERROR_501, template=cite.template.name),
                501
            )
    else:
        return make_response(format_page(ERROR_400, query=query), 400)

@_APP.route('/')
def _index():
    """
    Handle query if one is provided with the 's' parameter, otherwise
    show the main page.
    """
    query = request.args.get('s')
    if query:
        return _handle_query(query)
    else:
        return _APP.index_page

# Pregenerated Pages

@_APP.route('/sources')
def _sources():
    "return a page listing the citator's templates in a table"
    return _APP.sources_page

@_APP.route('/parser', methods= ['POST', 'GET'])
def _linker():
    "return a page that parses user-provided text"
    if request.method == 'GET':
        return format_page(PARSER_PAGE, given_text='', output='')
        
    given_text = escape(request.form['text'])
    if _APP.max_chars and len(given_text) > _APP.max_chars:
        return format_page(ERROR_501)
    citations = _APP.citator.list_cites(given_text)
    
    if not citations:
        return format_page(
            PARSER_PAGE,
            given_text=given_text,
            output="<p>Sorry, I couldn't find any citations in that.</p>"
        )
    
    output = insert_links(
        text = given_text,
        redundant_links = True,
        ignore_markup = False,
        citator = _APP.citator,
    )
    output = '<p>' + sub(r'\n+', '</p>\n<p>', output) + '</p>'
    
    return format_page(
        PARSER_PAGE,
        given_text=given_text,
        output=(
            '<h2 id="output">Output</h2>\n'
            + f'<form><div class="output-box">{output}</div></form>'
        ),
    )

# Static Files

@_APP.route('/logo.svg')
def _logo():
    return send_file('logo.svg')

@_APP.route('/style.css')
def _css():
    return send_file('style.css')

########################################################################
# Utility Functions
########################################################################

def _get_local_ip():
    "get local IP address. source: https://stackoverflow.com/a/28950776"
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(('10.255.255.255', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP

########################################################################
# Public Functions
########################################################################

def App(citator: Citator=Citator(), name: str='CiteURL'):
    """
    Return a flask application to implement the given citator. If called
    repeatedly, it will probably overwrite earlier instances.
    """
    # store variables in the app itself
    _APP.name = name
    _APP.index_page = format_page(INDEX_PAGE, name=name)
    _APP.citator = citator
    _APP.max_chars=400000 # limit query length
    _APP.sources_page = format_page(
        SOURCES_PAGE,
        intro=SOURCES_INTRO,
        table=sources_table(citator),
    )
    return _APP

def serve(app, localhost=True, port=53037):
    """
    Use gevent to host the app as a WSGI server at the given port. If
    not localhost, use _get_local_ip() to find the IP address to use.
    """ 
    from gevent import monkey
    from gevent.pywsgi import WSGIServer
    monkey.patch_all()
    
    if localhost:
        IP = 'localhost'
    else:
        IP = _get_local_ip()
    
    base_URL = f'http://{IP}:{port}'

    server = WSGIServer((IP, port), app)
    print(f'Now hosting citation lookup server at {base_URL}...')
    try:
        server.serve_forever()
    except KeyboardInterrupt:
        print("Stopped hosting citation lookup server.")



# python standard imports
import re
from copy import copy
from pathlib import Path

# third-party imports
from yaml import safe_load, safe_dump
# from appdirs import AppDirs        conditionally loaded later on

# internal imports
from .tokens import TokenType, StringBuilder
from .citation import Citation
from .authority import Authority, list_authorities
from .regex_mods import process_pattern, match_regexes

_DEFAULT_CITATOR = None

class Template:
    """
    A pattern to recognize a single kind of citation and extract
    information from it.
    """
    def __init__(
        self,
        name: str,
        tokens: dict[str, TokenType] = {},
        meta: dict[str, str] = {},
        patterns: list[str] = [],
        broad_patterns: list[str] = [],
        shortform_patterns: list[str] = [],
        idform_patterns: list[str] = [],
        name_builder: StringBuilder = None,
        URL_builder: StringBuilder = None,
        inherit_template = None,
    ):
        """
        Arguments:
            name: the name of this template
            
            tokens: The full dictionary of TokenTypes that citations from
                this template can contain. These must be listed in order
                from least-specific to most. For instance, the U.S.
                Constitution's template puts 'article' before 'section'
                before 'clause', because articles contain sections, and
                sections contain clauses.
            
            patterns: Patterns are essentially regexes to recognize
                recognize long-form citations to this template. However,
                wherever a token would appear in the regex, it should be
                replaced by the name of the token, enclosed in curly
                braces.
                
                Patterns are matched in the order that they are listed,
                so if there is a pattern that can only find a subset of
                tokens, it should be listed after the more-complete
                pattern so that the better match won't be precluded.
            
            broad_patterns: Same as `patterns`, except that they will
                only be used in contexts like search engines, where
                convenience is more important than avoiding false
                positive matches. When used, they will be used in
                addition to the normal patterns.
            
            shortform_patterns: Same as `patterns`, but these will only
                go into effect after a longform citation has been
                recognized. If a shortform pattern includes "same
                TOKEN_NAME" in curly braces, e.g. "{same volume}", the
                bracketed portion will be replaced with the exact text
                of the corresponding `raw_token` from the long-form
                citation.
            
            idform_patterns: Same as `shortform_patterns`, except that
                they will only be used to scan text until the next
                different citation occurs.
            
            URL_builder: `StringBuilder` to construct URLs for found
                citations
            
            name_builder: `StringBuilder` to construct canonical names
                of found citations
            
            meta: Optional metadata relating to this template. Patterns
                and StringBuilders can access metadata fields as if they
                were tokens, though fields can be overridden by tokens
                with the same name.
            
            inherit_template: another `Template` whose values this one
                should copy unless expressly overwritten.
        """
        kwargs = locals()
        for attr, default in {
            'name':               None,
            'tokens':             {},
            'patterns':           [],
            'broad_patterns':     [],
            'shortform_patterns': [],
            'idform_patterns':    [],
            'URL_builder':        None,
            'name_builder':       None,
            'meta':               {},
        }.items():
            if inherit_template and kwargs[attr] == default:
                value = inherit_template.__dict__.get(attr)
            elif attr.endswith('patterns') and not kwargs[attr]:
                value = []
            else:
                value = kwargs[attr]
            self.__dict__[attr] = value
        
        # update inherited StringBuilders with the correct metadata
        if inherit_template and self.meta:
            if self.URL_builder:
                self.URL_builder = copy(self.URL_builder)
                self.URL_builder.defaults = self.meta
            if self.name_builder:
                self.name_builder = copy(self.name_builder)
                self.name_builder.defaults = self.meta
        
        # use the template's metadata and tokens to make a dictionary
        # of replacements to insert into the regexes before compilation
        replacements = {k:str(v) for (k, v) in self.meta.items()}
        replacements.update({
            k:fr'(?P<{k}>{v.regex})(?!\w)'
            for (k,v) in self.tokens.items()
        })
        
        # compile the template's regexes and broad_regexes
        self.regexes = []
        self.broad_regexes = []
        for kind in ['regexes', 'broad_regexes']:
            if kind == 'broad_regexes':
                pattern_list = self.patterns + self.broad_patterns
                flags = re.I
            else:
                pattern_list = self.patterns
                flags = 0
            
            for p in pattern_list:
                pattern = process_pattern(
                    p,
                    replacements,
                    add_word_breaks=True
                )
                try:
                    regex = re.compile(pattern, flags)
                    self.__dict__[kind].append(regex)
                except re.error as e:
                    i = 'broad ' if kind == 'broad_regexes' else ''
                    raise re.error(
                        f'{self} template\'s {i}pattern "{pattern}" has '
                        f'an error: {e}'
                    )
        
        self._processed_shortforms = [
            process_pattern(p, replacements, add_word_breaks=True)
            for p in self.shortform_patterns
        ]
        self._processed_idforms = [
            process_pattern(p, replacements, add_word_breaks=True)
            for p in self.idform_patterns
        ]
    
    @classmethod
    def from_dict(cls, name: str, values: dict, inheritables: dict={}):
        """
        Return a template from a dictionary of values, like a dictionary
        created by parsing a template from YAML format.
        """
        values = {
            k.replace(' ', '_'):v
            for k,v in values.items()
        }
        
        # when pattern is listed in singular form,
        # replace it with a one-item list
        items = values.items()
        values = {}
        for key, value in items:
            if key.endswith('pattern'):
                values[key + 's'] = [value]
            else:
                values[key] = value
        
        # unrelated: when a single pattern is split
        # into a list (likely to take advantage of
        # YAML anchors), join it into one string
        for k,v in values.items():
            if not k.endswith('patterns'):
                continue
            elif v is None:
                values[k] = None
                continue
            for i, pattern in enumerate(v):
                if type(pattern) is list:
                    values[k][i] = ''.join(pattern)
        
        inherit = values.get('inherit')
        
        if inherit:
            values.pop('inherit')
            try:
                values['inherit_template'] = inheritables.get(inherit)
            except KeyError:
                raise KeyError(
                    f'Template "{name}" tried to inherit unknown '
                    + f'template "{inherit}"'
                )
        
        for key in ['name_builder', 'URL_builder']:
            data = values.get(key)
            if data:
                data['defaults'] = values.get('meta') or {}
                values[key] = StringBuilder.from_dict(data)
        values['tokens'] = {
            k: TokenType.from_dict(k, v)
            for k,v in values.get('tokens', {}).items()
        }
        return cls(name=name, **values)
    
    def to_dict(self) -> dict:
        "save this Template to a dictionary of values"
        output = {}
        if self.meta:
            output['meta'] = self.meta
        output['tokens'] = {
            k:v.to_dict() for k, v in self.tokens.items()
        }
        for key in ['patterns', 'shortform_patterns', 'idform_patterns']:
            value = self.__dict__.get(key)
            if not value:
                continue
            elif len(value) > 1:
                output[key] = value
            else: # de-pluralize lists that contain only one pattern
                output[key[:-1]] = value[0]
        for key in ['name_builder', 'URL_builder']:
            if self.__dict__.get(key):
                output[key] = self.__dict__[key].to_dict()
        
        spaced_output = {k.replace('_', ' '):v for k, v in output.items()}
        
        return spaced_output
    
    def to_yaml(self) -> str:
        "save this Template to a YAML string"
        return safe_dump(
            {self.name: self.to_dict()},
            sort_keys = False,
            allow_unicode = True,
        )
    
    def cite(self, text, broad: bool=True, span: tuple=(0,)) -> Citation:
        """
        Return the first citation that matches this template. If 'broad'
        is True, case-insensitive matching and broad regex patterns will
        be used. If no matches are found, return None.
        """
        regexes = self.broad_regexes if broad else self.regexes
        matches = match_regexes(text, regexes, span=span)
        for match in matches:
            try:
                return Citation(match, self)
            except SyntaxError: # invalid citation
                continue
        else:
            return None
    
    def list_longform_cites(self, text, broad: bool=False, span: tuple=(0,)):
        """
        Get a list of all long-form citations to this template found in
        the given text.
        """
        cites = []
        regexes = self.broad_regexes if broad else self.regexes
        for match in match_regexes(text, regexes, span=span):
            try:
                cites.append(Citation(match, self))
            except SyntaxError:
                continue
        return cites
    
    def __str__(self):
        return self.name
        
    def __repr__(self):
        return (
            f'Template(name="{self.name}"'
            + (f', tokens={self.tokens}' if self.tokens else '')
            + (f', meta={self.meta}' if self.meta else '')
            + (f', patterns={self.patterns}' if self.patterns else '')
            + (
                f', broad_patterns={self.broad_patterns}' 
                if self.broad_patterns else ''
            )
            + (
                f', shortform_patterns={self.shortform_patterns}'
                if self.shortform_patterns else ''
            )
            + (
                f', idform_patterns={self.idform_patterns}'
                if self.idform_patterns else ''
            )
            + (
                f', name_builder={self.name_builder}'
                if self.name_builder else ''
            )
            + (
                f', URL_builder={self.URL_builder}'
                if self.URL_builder else ''
            )
            + ')'
        )
    
    def __contains__(self, citation: Citation):
        return citation.template.name == self.name
    
    def __eq__(self, other_template):
        return repr(self) == repr(other_template)



class Citator:
    """
    A collection of citation templates, and the tools to match text
    against them en masse.
    
    Attributes:
        templates: a dictionary of citation templates that this citator
            will try to match against
    """
    
    def __init__(
        self,
        defaults = [
            'caselaw',
            'general federal law',
            'specific federal laws',
            'state law',
            'secondary sources',
        ],
        yaml_paths: list[str] = [],
        templates: dict[str, Template] = {},
    ):
        """
        Create a citator from any combination of CiteURL's default
        template sets (by default, all of them), plus any custom
        templates you want, either by pointing to custom YAML files or
        making Template objects at runtime.
        
        Arguments:
            defaults: names of files to load from the citeurl/templates
                folder. Each file contains one or more of CiteURL's
                built-in templates relevant to the given topic.
            yaml_paths: paths to custom YAML files to load templates
                from. These are loaded after the defaults, so they can
                inherit and/or overwrite them. If 
            templates: optional list of Template objects to load
                directly. These are loaded last, after the defaults and
                any yaml_paths.
        """
        self.templates = {}
        
        yamls_path = Path(__file__).parent.absolute() / 'templates'    
        for name in defaults or []:
            yaml_file = yamls_path / f'{name}.yaml'
            self.load_yaml(yaml_file.read_text())
        
        for path in yaml_paths:
            self.load_yaml(Path(path).read_text())
        self.templates.update(templates)
    
    @classmethod
    def from_yaml(cls, yaml: str):
        """
        Create a citator from scratch (i.e. without the default
        templates) by loading templates from the specified YAML string.
        """
        citator = cls(defaults=None)
        citator.load_yaml(yaml)
        return citator
    
    def to_yaml(self):
        "Save this citator to a YAML string to load later"
        yamls = [t.to_yaml() for t in self.templates.values()]
        return '\n\n'.join(yamls)
    
    def load_yaml(self, yaml: str):
        """
        Load templates from the given YAML, overwriting any existing
        templates with the same name.
        """
        for name, data in safe_load(yaml).items():
            self.templates[name] = Template.from_dict(
                name, data, inheritables=self.templates
            )
    
    def cite(self, text: str, broad: bool=True) -> Citation:
        """
        Check the given text against each of the citator's templates and
        return the first citation detected, or None.
        
        If broad is true, matching is case-insensitive and each
        template's broad regexes are used in addition to its normal
        regexes.
        """
        for template in self.templates.values():
            cite = template.cite(text, broad=broad)
            if cite:
                return cite
        else:
            return None
    
    def list_cites(
        self,
        text: str,
        id_breaks: re.Pattern = None,
    ) -> list[Citation]:
        """
        Find all citations in the given text, whether longform,
        shortform, or idform. They will be listed in order of
        appearance. If any two citations overlap, the shorter one will
        be deleted. 
        
        Wherever the id_breaks pattern appears, it will interrupt chains
        of id-form citations. This is helpful for handling unrecognized
        citations that would otherwise cause CiteURL's notion of "id."
        to get out of sync with what the text is talking about.
        """
        # first get a list of all long and shortform (not id.) citations
        longforms = []
        for template in self.templates.values():
            longforms += template.list_longform_cites(text)

        shortforms = []
        for citation in longforms:
            shortforms += citation.get_shortform_cites()

        citations = longforms + shortforms
        _sort_and_remove_overlaps(citations)
        
        # Figure out where to interrupt chains of idform citations,
        # i.e. anywhere a longform or shortform citation starts, plus
        # the start of any substring that matches the id_breaks pattern
        breakpoints = [c.span[0] for c in citations]
        if id_breaks:
            breakpoints += [
                match.span()[0] for match in
                id_breaks.finditer(text)
            ]
        breakpoints = sorted(set(breakpoints))
        
        # for each cite, look for idform citations until the next cite
        # or until the next breakpoint
        idforms = []
        for cite in citations:
            # find the next relevant breakpoint, and delete any
            # breakpoints that are already behind the current citation
            for i, breakpoint in enumerate(breakpoints):
                if breakpoint >= cite.span[1]:
                    breakpoints = breakpoints[i:]
                    break
            try:
                breakpoint = breakpoints[0]
            except IndexError:
                breakpoint = None
            
            # find the first idform reference to the citation, then the
            # first idform reference to that idform, and so on, until
            # the breakpoint
            idform = cite.get_idform_cite(until_index=breakpoint)
            while idform:
                idforms.append(idform)
                idform = idform.get_idform_cite(until_index=breakpoint)
        
        citations += idforms
        _sort_and_remove_overlaps(citations)
        return citations
    
    def list_authorities(
        self,
        text: str,
        ignored_tokens = ['subsection', 'clause', 'pincite', 'paragraph'],
        known_authorities: list = [],
        sort_by_cites: bool = True,
        id_breaks: re.Pattern = None,
    ) -> list[Authority]:
        """
        Find each distinct authority mentioned in the given text, and 
        return Authority objects whose `citations` attribute lists the
        references to each.
        
        Arguments:
            text: The string to be scanned for citations
            ignored_tokens: the names of tokens whose values are
                irrelevant to whether the citation matches an authority,
                because they  just designate portions within a single
                authority
            sort_by_cites: Whether to sort the resulting list of
                authorities by the number of citations to each one
        """
        cites = self.list_cites(text, id_breaks=id_breaks)
        return list_authorities(
            cites,
            ignored_tokens = ignored_tokens,
            known_authorities = known_authorities,
            sort_by_cites = sort_by_cites,
        )        
    
    def insert_links(
        self,
        text: str,
        attrs: dict = {'class': 'citation'},
        add_title: bool = True,
        URL_optional: bool = False,
        redundant_links: bool = True,
        id_breaks: re.Pattern = None,
        ignore_markup: bool = True,
    ) -> str:
        """
        Scan a text for citations, and return a text with each citation
        converted to a hyperlink.
        
        Arguments:
            text: the string to scan for citations.
            attrs: various HTML link attributes to give to each link
            add_title: whether to use citation.name for link titles
            URL_optional: whether to insert a hyperlink even when the
                citation does not have an associated URL
            redundant_links: whether to insert a hyperlink if it would
                point to the same URL as the previous link
            id_breaks: wherever this regex appears, interrupt chains of
                "Id."-type citations.
            ignore_markup: whether to preprocess and postprocess the
                text so that CiteURL can detect citations even when
                they contain inline markup, like "<i>Id.</i> at 32"
        
        Returns:
            text, with an HTML `a` element for each citation. 
        """
        
        # pull out all the inline HTML tags, e.g. <b>,
        # so they don't interfere with citation matching
        if ignore_markup:
            text, stored_tags = _strip_inline_tags(text)
        
        cite_offsets = []
        running_offset = 0
        
        last_URL = None
        for cite in self.list_cites(text, id_breaks = id_breaks):
            attrs['href'] = cite.URL
            if not cite.URL and not URL_optional:
                continue
            if not redundant_links and cite.URL == last_URL:
                continue
            if add_title:
                attrs['title'] = cite.name
            
            attr_str = ''.join([
                f' {k}="{v}"'
                for k, v in attrs.items() if v
            ])
            link = f'<a{attr_str}>{cite.text}</a>'
            
            cite_offset = len(link) - len(cite.text)
                    
            cite_offsets.append((
                cite.span[0], # beginning of citation
                cite_offset,
                cite.text,
            ))
            
            span = (
                cite.span[0] + running_offset,
                cite.span[1] + running_offset
            )
            
            text = text[:span[0]] + link + text[span[1]:]
            
            running_offset += cite_offset
            last_URL = cite.URL
        
        if ignore_markup:
            running_offset = 0
            for tag in stored_tags:
                temp_offset = 0
                while len(cite_offsets) > 0:
                    # only offset by a cite if the tag
                    # is after the cite start
                    if tag[1] >= cite_offsets[0][0]:
                        offset = cite_offsets[0]
                        # check if the tag is after the cite end
                        tag_start = tag[1]
                        cite_end = offset[0] + len(offset[2])
                        
                        if tag_start >= cite_end:
                            running_offset += offset[1]
                            cite_offsets.pop(0)
                        # otherwise, don't offset by the length of
                        # the cite's closing </a> tag (length of 4)
                        else:
                            temp_offset = offset[1] - 4
                            break
                    else:
                        break
                tag_pos = tag[1] + running_offset + temp_offset
                
                text = text[:tag_pos] + tag[0] + text[tag_pos:]
                
                running_offset += tag[2]
        
        return text
    
    def __iter__(self):
        return self.templates.values().__iter__()
    
    def __getitem__(self, key):
        return self.templates[key]
    
    def __setitem__(self, key, value):
        self.templates[key] = value
    
    def __eq__(self, other_citator):
        return self.templates == other_citator.templates

########################################################################
# PUBLIC FUNCTIONS
########################################################################

def cite(
    text: str,
    broad: bool = True,
    citator: Citator = None,
) -> Citation:
    """
    Convenience function to find a single citation in text, or None. See
    Citator.cite() for more info.
    """
    citator = citator or _get_default_citator()
    return citator.cite(text, broad=broad)

def list_cites(text, citator: Citator = None, id_breaks=None):
    """
    Convenience function to list all citations in a text. For more info,
    see Citator.list_cites().
    """
    citator = citator or _get_default_citator()
    return citator.list_cites(text, id_breaks=id_breaks)

def insert_links(
    text: str,
    attrs: dict = {'class': 'citation'},
    add_title: bool = True,
    URL_optional: bool = False,
    redundant_links: bool = True,
    id_breaks: re.Pattern = None,
    ignore_markup: bool = True,
    citator: Citator = None,
):
    """
    Convenience function to hyperlink all citations in a text. For more
    info, see Citator.insert_links().
    """
    citator = citator or _get_default_citator()
    return citator.insert_links(
        text = text,
        attrs = attrs,
        add_title = add_title,
        redundant_links = redundant_links,
        id_breaks = id_breaks,
        ignore_markup = ignore_markup,
    )

########################################################################
# INTERNAL FUNCTIONS
########################################################################

def _sort_and_remove_overlaps(citations: list[Citation]):
    """
    For a given list of citations found in the same text, sort them by
    their order of appearance. When two citations overlap, the shorter
    one will be deleted. The list is modified in place.
    """
    citations.sort(key=lambda x: x.span[0])
    i = 1
    while i < len(citations):
        if citations[i].span[0] < citations[i-1].span[1]:
            if len(citations[i-1]) > len(citations[i]):
                citations.pop(i)
            else:
                citations.pop(i-1)
        else:
            i += 1

def _get_default_citator():
    """
    Instantiate a citator if needed, and reuse it otherwise. If appdirs
    is installed, load default templates from your config directory
    """
    global _DEFAULT_CITATOR
    if _DEFAULT_CITATOR:
        return _DEFAULT_CITATOR
    # load custom templates from config dir, if possible
    try:
        from appdirs import AppDirs
        _appdirs = AppDirs('citeurl', 'raindrum')
        _user_config_dir = Path(_appdirs.user_config_dir)
        user_templates = set([
            file for file in _user_config_dir.iterdir()
            if file.suffix.lower() in ['.yaml', '.yml']
        ])
    except ImportError:
        user_templates = {}
    _DEFAULT_CITATOR = Citator(yaml_paths=user_templates)
    return _DEFAULT_CITATOR

def _strip_inline_tags(text: str) -> tuple[str, list[tuple]]:
    inline_tag_regex = '|'.join([
        'a', 'abbr', 'acronym', 'b', 'bdo', 'big', 'br', 'button', 'cite',
        'code', 'dfn', 'em', 'i', 'img', 'input', 'kbd', 'label', 'map',
        'object', 'output', 'q', 'samp', 'script', 'select', 'small', 'span',
        'strong', 'sub', 'sup', 'textarea', 'time', 'tt', 'var', 
    ])
    stored_tags = []
    offset = 0
    inline_tag_regex = f'</?({inline_tag_regex})(>| .+?>)'
    def store_tag(match):
        nonlocal offset
        tag_text = match.group(0)
        tag_length = len(tag_text)
        tag_start = match.span()[0] - offset
        stored_tags.append((
            tag_text,
            tag_start,
            tag_length,
        ))
        offset += tag_length
        return ''
    text = re.sub(inline_tag_regex, store_tag, text)
    return text, stored_tags



import setuptools

with open('README.md', 'r') as f:
    readme = f.read()

setuptools.setup(
    name = 'citeurl',
    version = '11.2.0',
    description = 'an extensible tool to process legal citations in text',
    author = 'Simon Raindrum Sherred',
    author_email = 'simonraindrum@gmail.com',
    license = 'MIT',
    long_description = readme,
    long_description_content_type = "text/markdown",
    url="https://raindrum.github.io/citeurl",
    packages = setuptools.find_packages(),
    entry_points = {
        'console_scripts': ['citeurl=citeurl.cli:main'],
        'markdown.extensions': ['citeurl=citeurl.mdx:CiteURLExtension'],
    },
    include_package_data = True,
    install_requires = ['pyyaml'],
    extras_require = {
        'web': ['flask'],
        'config': ['appdirs'],
        'full': ['flask', 'appdirs'],
    },
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Legal Industry',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Text Processing :: Filters',
        'Topic :: Text Processing :: Markup :: HTML',
        'Topic :: Text Processing :: Markup :: Markdown',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



# python standard imports
from typing import Iterable
import re

# internal imports
from .regex_mods import process_pattern, match_regexes

BASIC_ID_REGEX = re.compile(r'(?<!\w)[Ii](bi)?d\.(?!=\w)')

class Citation:
    """
    A legal reference found in text.
    
    Attributes:
        tokens: dictionary of the values that define this citation, such
            as its volume and page number, or its title, section, and
            subsection, etc
        
        URL: the location, if any, where this citation can be found
            online, defined by the template's URL_builder
        
        name: a uniform, human-readable representation of this citation,
            written by the template's name_builder
        
        text: the actual text of this citation as found in the source
            text
        
        source_text: the full text that this citation was found in
        
        template: the template whose regexes found this citation or its
            parent
        
        parent: the earlier citation, if any, that this citation is a
            shortform or idform child of
        
        raw_tokens: dictionary of tokens as captured in the original
            regex match, before normalization. Note that for child
            citations, raw_tokens will include any raw_tokens inferred
            from the parent citation.
        
        idform_regexes: list of regex pattern objects to find child
            citations later in the text, valid until the next different
            citation appears.
        
        shortform_regexes: list of regex pattern objects to find
            child citations anywhere in the subsequent text
    """
    
    def __init__(
        self,
        match: re.match,
        template,
        parent = None,
    ):
        self.match = match
        self.text = match.group(0)
        self.source_text = match.string
        self.span = match.span()
        self.template = template
        self.parent = parent
        self.tokens = {}
        self.raw_tokens = match.groupdict()
        
        # copy raw_tokens (in order) from the parent citation, but
        # stop at the first one that the child citation overwrites
        if parent:
            merged_tokens = {}
            for k in template.tokens.keys():
                if self.raw_tokens.get(k):
                    merged_tokens.update(self.raw_tokens)
                    break
                else:
                    merged_tokens[k] = parent.raw_tokens.get(k)
            self.raw_tokens = merged_tokens
        
        # normalize raw_tokens to get consistent token values across
        # differently-formatted citations to the same source.
        # This will raise a SyntaxError if a mandatory edit fails
        for name, ttype in template.tokens.items():
            value = self.raw_tokens.get(name)
            self.tokens[name] = ttype.normalize(value)
        
        # Finally, compile the citation's idform and shortform regexes.
        # To avoid unneccessary work, first try to copy regexes from the
        # parent citation if applicable.
        
        if parent and parent.raw_tokens == self.raw_tokens:
        # then we can safely copy the parent's regexes to the child
            self.idform_regexes = parent.idform_regexes
            self.shortform_regexes = parent.shortform_regexes
            return
        
        # otherwise we'll need to compile new shortform regexes,
        # but we can still copy some of them from the parent
        
        kwargs = {
            'replacements': self.raw_tokens,
            'token_prefix': 'same',
        }
        if parent:
        # we can copy regexes, but only if they do not reference a
        # specific value from the citation, e.g. {same volume}.
            self.shortform_regexes = [
                (
                    re.compile(process_pattern(pattern, **kwargs))
                    if '{same ' in pattern else parent.shortform_regexes[i]
                )
                for i, pattern in enumerate(template._processed_shortforms)
            ]
            
            self.idform_regexes = [
                (
                    re.compile(process_pattern(pattern, **kwargs))
                    if '{same ' in pattern else parent.idform_regexes[i]
                )
                for i, pattern in enumerate(template._processed_idforms)
            ]
            
        else: # compile all-new idforms and shortforms
            self.shortform_regexes = [
                re.compile(process_pattern(pattern, **kwargs))
                for pattern in self.template._processed_shortforms
            ]
            self.idform_regexes = [
                re.compile(process_pattern(pattern, **kwargs))
                for pattern in self.template._processed_idforms
            ]
        self.idform_regexes.append(BASIC_ID_REGEX)
    
    @property
    def URL(self) -> str:
        if self.template.URL_builder:
            url =  self.template.URL_builder(self.tokens)
            if url:
                url = url.replace(' ', '%20')
        else:
            url = None
        return url
    
    @property
    def name(self) -> str:
        if self.template.name_builder:
            return self.template.name_builder(self.tokens)
        else:
            return None
    
    def get_shortform_cites(self) -> Iterable:
        keep_trying = True
        span_start = self.span[1]
        while keep_trying:
            try:
                match = next(match_regexes(
                    regexes=self.shortform_regexes,
                    text=self.source_text,
                    span=(span_start,),
                ))
                span_start = match.span()[1]
                try:
                    yield Citation(
                        match=match,
                        template=self.template,
                        parent=self,
                    )
                except SyntaxError: # it's an invalid citation
                    pass
            except StopIteration:
                keep_trying = False
    
    def get_idform_cite(self, until_index: int=None):
        try:
            match = next(match_regexes(
                regexes = self.idform_regexes,
                text = self.source_text,
                span = (self.span[1], until_index)
            ))
            return Citation(match=match, template=self.template, parent=self)
        except StopIteration:
            return None
        except SyntaxError:
            return None
    
    def get_next_child(self, span: tuple=None):
        try:
            match = next(match_regexes(
                regexes = self.shortform_regexes + self.idform_regexes,
                text = self.source_text,
                span = span if span else (self.span[1], ),
            ))
            return Citation(match=match, template=self.template, parent=self)
        except StopIteration:
            return None
    
    def __str__(self):
        return str(self.text)
    
    def __repr__(self):
        return str(self.text)
        return (
            f'Citation(match={self.match}, template={repr(self.template)}'
            + (f', parent={repr(self.parent)}' if self.parent else '')
        )
    
    def __contains__(self, other_cite):
        """
        Returns True if both citations are from templates with the same
        name, and the only difference between their tokens is that the
        other one has a more specific (i.e. higher-indexed) token than
        any of this one's. Severable tokens are considered a match if
        the other token's value *starts with* this one's.
        """
        if (
            other_cite.template.name != self.template.name
            or other_cite.tokens == self.tokens
        ):
            return False
        for key, value in self.tokens.items():
            if value and other_cite.tokens.get(key) != value:
                if (
                    self.template.tokens[key].severable
                    and other_cite.tokens[key]
                    and other_cite.tokens[key].startswith(value)
                ):
                    continue
                else:
                    return False
        else:
            return True
    
    def __eq__(self, other_cite):
        """
        Returns True if both citations are from templates with the same
        name, and they have the exact same token values.
        """
        return (
            other_cite.template.name == self.template.name
            and other_cite.tokens == self.tokens
        )
    
    def __len__(self):
        return len(self.text)

